@@ -1,4 +1,4 @@
-from __future__ import division
+from __future__ import division, absolute_import, print_function
 
 __copyright__ = "Copyright (C) 2009 Andreas Kloeckner"
 
@@ -22,8 +22,11 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 """
 
+from six.moves import range
+
 import numpy as np
 import numpy.linalg as la
+import pytest
 
 import pyopencl as cl
 import pyopencl.array as cl_array
@@ -39,6 +42,12 @@ else:
     faulthandler.enable()
 
 
+def _skip_if_pocl(plat, msg='unsupported by pocl'):
+    if plat.vendor == "The pocl project":
+        import pytest
+        pytest.skip(msg)
+
+
 def test_get_info(ctx_factory):
     ctx = ctx_factory()
     device, = ctx.devices
@@ -47,16 +56,26 @@ def test_get_info(ctx_factory):
     failure_count = [0]
 
     pocl_quirks = [
-            (cl.Buffer, cl.mem_info.OFFSET),
+        (cl.Buffer, cl.mem_info.OFFSET),
+        (cl.Program, cl.program_info.BINARIES),
+        (cl.Program, cl.program_info.BINARY_SIZES),
+    ]
+    if ctx._get_cl_version() >= (1, 2) and cl.get_cl_header_version() >= (1, 2):
+        pocl_quirks.extend([
             (cl.Program, cl.program_info.KERNEL_NAMES),
             (cl.Program, cl.program_info.NUM_KERNELS),
-            ]
-    CRASH_QUIRKS = [
+        ])
+    CRASH_QUIRKS = [  # noqa
             (("NVIDIA Corporation", "NVIDIA CUDA",
                 "OpenCL 1.0 CUDA 3.0.1"),
                 [
                     (cl.Event, cl.event_info.COMMAND_QUEUE),
                     ]),
+            (("NVIDIA Corporation", "NVIDIA CUDA",
+                "OpenCL 1.2 CUDA 7.5"),
+                [
+                    (cl.Buffer, getattr(cl.mem_info, "USES_SVM_POINTER", None)),
+                    ]),
             (("The pocl project", "Portable Computing Language",
                 "OpenCL 1.2 pocl 0.8-pre"),
                     pocl_quirks),
@@ -66,22 +85,29 @@ def test_get_info(ctx_factory):
             (("The pocl project", "Portable Computing Language",
                 "OpenCL 1.2 pocl 0.9-pre"),
                 pocl_quirks),
+            (("The pocl project", "Portable Computing Language",
+                "OpenCL 1.2 pocl 0.9"),
+                pocl_quirks),
+            (("The pocl project", "Portable Computing Language",
+                "OpenCL 1.2 pocl 0.10-pre"),
+                pocl_quirks),
+            (("The pocl project", "Portable Computing Language",
+                "OpenCL 1.2 pocl 0.10"),
+                pocl_quirks),
             (("Apple", "Apple",
-                "OpenCL 1.2 (Apr 25 2013 18:32:06)"),
+                "OpenCL 1.2"),
                 [
                     (cl.Program, cl.program_info.SOURCE),
                     ]),
             ]
-    QUIRKS = []
-
-    plat_quirk_key = (
-            platform.vendor,
-            platform.name,
-            platform.version)
+    QUIRKS = []  # noqa
 
     def find_quirk(quirk_list, cl_obj, info):
-        for entry_plat_key, quirks in quirk_list:
-            if entry_plat_key == plat_quirk_key:
+        for (vendor, name, version), quirks in quirk_list:
+            if (
+                    vendor == platform.vendor
+                    and name == platform.name
+                    and platform.version.startswith(version)):
                 for quirk_cls, quirk_info in quirks:
                     if (isinstance(cl_obj, quirk_cls)
                             and quirk_info == info):
@@ -165,6 +191,7 @@ def test_get_info(ctx_factory):
                 try_attr_form=False)
 
     # crashes on intel...
+    # and pocl does not support CL_ADDRESS_CLAMP
     if device.image_support and platform.vendor not in [
             "Intel(R) Corporation",
             "The pocl project",
@@ -218,6 +245,7 @@ def test_int_ptr(ctx_factory):
     do_test(a_buf)
 
     # crashes on intel...
+    # and pocl does not support CL_ADDRESS_CLAMP
     if device.image_support and platform.vendor not in [
             "Intel(R) Corporation",
             "The pocl project",
@@ -242,12 +270,6 @@ def test_invalid_kernel_names_cause_failures(ctx_factory):
         { a[get_global_id(0)] *= 2; }
         """).build()
 
-    if ctx.devices[0].platform.vendor == "The pocl project":
-        # https://bugs.launchpad.net/pocl/+bug/1184464
-
-        import pytest
-        pytest.skip("pocl doesn't like invalid kernel names")
-
     try:
         prg.sam
         raise RuntimeError("invalid kernel name did not cause error")
@@ -329,12 +351,7 @@ def test_image_2d(ctx_factory):
     if "Intel" in device.vendor and "31360.31426" in device.version:
         from pytest import skip
         skip("images crashy on %s" % device)
-    if "pocl" in device.platform.vendor and (
-            "0.8" in device.platform.version or
-            "0.9" in device.platform.version
-            ):
-        from pytest import skip
-        skip("images crashy on %s" % device)
+    _skip_if_pocl(device.platform, 'pocl does not support CL_ADDRESS_CLAMP')
 
     prg = cl.Program(context, """
         __kernel void copy_image(
@@ -406,6 +423,7 @@ def test_image_3d(ctx_factory):
     if device.platform.vendor == "Intel(R) Corporation":
         from pytest import skip
         skip("images crashy on %s" % device)
+    _skip_if_pocl(device.platform, 'pocl does not support CL_ADDRESS_CLAMP')
 
     prg = cl.Program(context, """
         __kernel void copy_image_plane(
@@ -487,21 +505,22 @@ def test_copy_buffer(ctx_factory):
 
 
 def test_mempool(ctx_factory):
-    from pyopencl.tools import MemoryPool, CLAllocator
+    from pyopencl.tools import MemoryPool, ImmediateAllocator
 
     context = ctx_factory()
+    queue = cl.CommandQueue(context)
 
-    pool = MemoryPool(CLAllocator(context))
-    queue = []
+    pool = MemoryPool(ImmediateAllocator(queue))
+    alloc_queue = []
 
     e0 = 12
 
     for e in range(e0-6, e0-4):
         for i in range(100):
-            queue.append(pool.allocate(1 << e))
-            if len(queue) > 10:
-                queue.pop(0)
-    del queue
+            alloc_queue.append(pool.allocate(1 << e))
+            if len(alloc_queue) > 10:
+                alloc_queue.pop(0)
+    del alloc_queue
     pool.stop_holding()
 
 
@@ -593,6 +612,257 @@ def test_can_build_binary(ctx_factory):
     foo.build()
 
 
+def test_enqueue_barrier_marker(ctx_factory):
+    ctx = ctx_factory()
+    _skip_if_pocl(ctx.devices[0].platform, 'pocl crashes on enqueue_barrier')
+    queue = cl.CommandQueue(ctx)
+    cl.enqueue_barrier(queue)
+    evt1 = cl.enqueue_marker(queue)
+    evt2 = cl.enqueue_marker(queue, wait_for=[evt1])
+    cl.enqueue_barrier(queue, wait_for=[evt1, evt2])
+
+
+def test_wait_for_events(ctx_factory):
+    ctx = ctx_factory()
+    queue = cl.CommandQueue(ctx)
+    evt1 = cl.enqueue_marker(queue)
+    evt2 = cl.enqueue_marker(queue)
+    cl.wait_for_events([evt1, evt2])
+
+
+def test_unload_compiler(platform):
+    if (platform._get_cl_version() < (1, 2) or
+            cl.get_cl_header_version() < (1, 2)):
+        from pytest import skip
+        skip("clUnloadPlatformCompiler is only available in OpenCL 1.2")
+    _skip_if_pocl(platform, 'pocl does not support unloading compiler')
+    if platform.vendor == "Intel(R) Corporation":
+        from pytest import skip
+        skip("Intel proprietary driver does not support unloading compiler")
+    cl.unload_platform_compiler(platform)
+
+
+def test_enqueue_task(ctx_factory):
+    ctx = ctx_factory()
+    queue = cl.CommandQueue(ctx)
+    mf = cl.mem_flags
+
+    prg = cl.Program(ctx, """
+    __kernel void
+    reverse(__global const float *in, __global float *out, int n)
+    {
+        for (int i = 0;i < n;i++) {
+            out[i] = in[n - 1 - i];
+        }
+    }
+    """).build()
+    knl = prg.reverse
+
+    n = 100
+    a = np.random.rand(n).astype(np.float32)
+    b = np.empty_like(a)
+
+    buf1 = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=a)
+    buf2 = cl.Buffer(ctx, mf.WRITE_ONLY, b.nbytes)
+
+    knl.set_args(buf1, buf2, np.int32(n))
+    cl.enqueue_task(queue, knl)
+
+    cl.enqueue_copy(queue, b, buf2).wait()
+    assert la.norm(a[::-1] - b) == 0
+
+
+def test_platform_get_devices(ctx_factory):
+    ctx = ctx_factory()
+    platform = ctx.devices[0].platform
+
+    if platform.name == "Apple":
+        pytest.xfail("Apple doesn't understand all the values we pass "
+                "for dev_type")
+
+    dev_types = [cl.device_type.ACCELERATOR, cl.device_type.ALL,
+                 cl.device_type.CPU, cl.device_type.DEFAULT, cl.device_type.GPU]
+    if (platform._get_cl_version() >= (1, 2) and
+            cl.get_cl_header_version() >= (1, 2)
+            and not platform.name.lower().startswith("nvidia")):
+        dev_types.append(cl.device_type.CUSTOM)
+
+    for dev_type in dev_types:
+        print(dev_type)
+        devs = platform.get_devices(dev_type)
+        if dev_type in (cl.device_type.DEFAULT,
+                        cl.device_type.ALL,
+                        getattr(cl.device_type, 'CUSTOM', None)):
+            continue
+        for dev in devs:
+            assert dev.type & dev_type == dev_type
+
+
+def test_user_event(ctx_factory):
+    ctx = ctx_factory()
+    if (ctx._get_cl_version() < (1, 1) and
+            cl.get_cl_header_version() < (1, 1)):
+        from pytest import skip
+        skip("UserEvent is only available in OpenCL 1.1")
+
+    if ctx.devices[0].platform.name == "Portable Computing Language":
+        # https://github.com/pocl/pocl/issues/201
+        pytest.xfail("POCL's user events don't work right")
+
+    status = {}
+
+    def event_waiter1(e, key):
+        e.wait()
+        status[key] = True
+
+    def event_waiter2(e, key):
+        cl.wait_for_events([e])
+        status[key] = True
+
+    from threading import Thread
+    from time import sleep
+    evt = cl.UserEvent(ctx)
+    Thread(target=event_waiter1, args=(evt, 1)).start()
+    sleep(.05)
+    if status.get(1, False):
+        raise RuntimeError('UserEvent triggered before set_status')
+    evt.set_status(cl.command_execution_status.COMPLETE)
+    sleep(.05)
+    if not status.get(1, False):
+        raise RuntimeError('UserEvent.wait timeout')
+    assert evt.command_execution_status == cl.command_execution_status.COMPLETE
+
+    evt = cl.UserEvent(ctx)
+    Thread(target=event_waiter2, args=(evt, 2)).start()
+    sleep(.05)
+    if status.get(2, False):
+        raise RuntimeError('UserEvent triggered before set_status')
+    evt.set_status(cl.command_execution_status.COMPLETE)
+    sleep(.05)
+    if not status.get(2, False):
+        raise RuntimeError('cl.wait_for_events timeout on UserEvent')
+    assert evt.command_execution_status == cl.command_execution_status.COMPLETE
+
+
+def test_buffer_get_host_array(ctx_factory):
+    ctx = ctx_factory()
+    mf = cl.mem_flags
+
+    host_buf = np.random.rand(25).astype(np.float32)
+    buf = cl.Buffer(ctx, mf.READ_WRITE | mf.USE_HOST_PTR, hostbuf=host_buf)
+    host_buf2 = buf.get_host_array(25, np.float32)
+    assert (host_buf == host_buf2).all()
+    assert (host_buf.__array_interface__['data'][0] ==
+            host_buf.__array_interface__['data'][0])
+    assert host_buf2.base is buf
+
+    buf = cl.Buffer(ctx, mf.READ_WRITE | mf.ALLOC_HOST_PTR, size=100)
+    try:
+        host_buf2 = buf.get_host_array(25, np.float32)
+        assert False, ("MemoryObject.get_host_array should not accept buffer "
+                       "without USE_HOST_PTR")
+    except cl.LogicError:
+        pass
+
+    host_buf = np.random.rand(25).astype(np.float32)
+    buf = cl.Buffer(ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=host_buf)
+    try:
+        host_buf2 = buf.get_host_array(25, np.float32)
+        assert False, ("MemoryObject.get_host_array should not accept buffer "
+                       "without USE_HOST_PTR")
+    except cl.LogicError:
+        pass
+
+
+def test_program_valued_get_info(ctx_factory):
+    ctx = ctx_factory()
+
+    prg = cl.Program(ctx, """
+    __kernel void
+    reverse(__global float *out)
+    {
+        out[get_global_id(0)] *= 2;
+    }
+    """).build()
+
+    knl = prg.reverse
+
+    assert knl.program == prg
+    knl.program.binaries[0]
+
+
+def test_event_set_callback(ctx_factory):
+    ctx = ctx_factory()
+    queue = cl.CommandQueue(ctx)
+
+    if ctx._get_cl_version() < (1, 1):
+        pytest.skip("OpenCL 1.1 or newer required fro set_callback")
+
+    a_np = np.random.rand(50000).astype(np.float32)
+    b_np = np.random.rand(50000).astype(np.float32)
+
+    got_called = []
+
+    def cb(status):
+        got_called.append(status)
+
+    mf = cl.mem_flags
+    a_g = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=a_np)
+    b_g = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=b_np)
+
+    prg = cl.Program(ctx, """
+    __kernel void sum(__global const float *a_g, __global const float *b_g,
+        __global float *res_g) {
+      int gid = get_global_id(0);
+      res_g[gid] = a_g[gid] + b_g[gid];
+    }
+    """).build()
+
+    res_g = cl.Buffer(ctx, mf.WRITE_ONLY, a_np.nbytes)
+
+    uevt = cl.UserEvent(ctx)
+
+    evt = prg.sum(queue, a_np.shape, None, a_g, b_g, res_g, wait_for=[uevt])
+
+    evt.set_callback(cl.command_execution_status.COMPLETE, cb)
+
+    uevt.set_status(cl.command_execution_status.COMPLETE)
+
+    queue.finish()
+
+    # yuck
+    from time import sleep
+    sleep(0.1)
+
+    assert got_called
+
+
+def test_global_offset(ctx_factory):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    prg = cl.Program(context, """
+        __kernel void mult(__global float *a)
+        { a[get_global_id(0)] *= 2; }
+        """).build()
+
+    n = 50
+    a = np.random.rand(n).astype(np.float32)
+
+    queue = cl.CommandQueue(context)
+    mf = cl.mem_flags
+    a_buf = cl.Buffer(context, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=a)
+
+    step = 10
+    for ofs in range(0, n, step):
+        prg.mult(queue, (step,), None, a_buf, global_offset=(ofs,))
+
+    a_2 = np.empty_like(a)
+    cl.enqueue_copy(queue, a_2, a_buf)
+
+    assert (a_2 == 2*a).all()
+
+
 if __name__ == "__main__":
     # make sure that import failures get reported, instead of skipping the tests.
     import pyopencl  # noqa
